/* -*- Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 2000 Regents of the University of New Hampshire.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the MASH Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /opt/src/mash/repository/vint/ns-2/cell.h,v 1.0 2000/08/15 10:37 sow Exp $
 *
 * Contributed by Sowmya Manjanatha, Dept. of Computer Science, UNH, August 2000 
 */

#include "aal.h"
#include "ip.h"
#include <list>
#include <stl.h>
#include <vector>

static class AAL5Class : public TclClass {
public:
	AAL5Class() : TclClass("AAL5") {}
	TclObject* create(int, const char*const*) {
		return (new AAL5());
		
	}	
} class_aal5;



Packet **AAL5::segment(Packet *p) {

	int size = hdr_cmn::access(p)->size_ + 8;
	
	aal_uid = 0;
	int x = size % 48;
        size += (48 - x);

	reassembly->tx_cell_cnt = size/48;
	Packet *cell[2000];

	for (int i = 0; i < reassembly->tx_cell_cnt; i++) {
		cell[i] = Packet::alloc();
                hdr_atm *atmh = hdr_atm::access(cell[i]);
                atmh->packet = p;
		initpkt(cell[i]);
		atmh->vpi = reassembly->vpi;
		atmh->vci = reassembly->vci;
		
		strcpy(atmh->switcher_,"");
		cell[i]->set_atm();

		if (i == (reassembly->tx_cell_cnt - 1)) {
			atmh->type = LAST_CELL;
		}
		else 
			atmh->type = 0;
	}
	     
     return cell;

}

int AAL5::command(int argc, const char*const* argv) {

       return (TclObject::command(argc,argv));

}

void AAL5::set_parameters(Reassembly *r, string type) {

	reassembly = r;
	aal_type_ = type;
}

void AAL5::initpkt(Packet *p) 
{
	char buffer[20];

	hdr_cmn* ch = hdr_cmn::access(p);
	hdr_atm *atmh = hdr_atm::access(p);
	hdr_ip *iph = hdr_ip::access(p);

	hdr_ip *orig = hdr_ip::access(atmh->packet);


	ch->uid() = aal_uid++;

	ch->ptype() = PT_ATM;
	ch->size() = CELLSIZE;

	double local_time = Scheduler::instance().clock();

	ch->timestamp() = (u_int32_t)(SAMPLERATE * local_time);
	ch->iface() = UNKN_IFACE.value(); // from packet.h (agent is local)
	ch->direction() = hdr_cmn::NONE;
	ch->ref_count() = 0;	/* reference count */
	ch->error() = 0;	/* pkt not corrupt to start with */

	//required to find path of simulation
	iph->ttl() = 1000000;
	iph->saddr() = orig->saddr();
	iph->sport() = orig->sport();
	iph->daddr() = orig->daddr();
	iph->dport() = orig->dport();

	
}

void AAL5::increment() {

	reassembly->tx_cell_cnt++;
	
}

void AAL5::rx_increment() {

	reassembly->rx_cell_cnt++;
	
}

bool AAL5::cellcount_matches() {
	
	if(reassembly->tx_cell_cnt == reassembly->rx_cell_cnt) 
		return true;
	   
	return false;
}
	   
	
Packet *AAL5::build_frm(Packet *orig) {


	hdr_atm *temp = hdr_atm::access(orig);
	//general elements of an ATM packet (Cell)
	Packet *p = Packet::alloc();
	hdr_atm *atmh = hdr_atm::access(p);
	atmh->packet = orig;
	initpkt(p);
	atmh->vpi = reassembly->vpi;
	atmh->vci = reassembly->vci;
	atmh->type = 6;         //indicates RM cell
	p->set_atm();
	atmh->packet = temp->packet;

	//parameters specific to FRM cell

	string rm = string("00000001");   //protocol identifier for ABR

	//Message type, length = 1 octet
	(rm).append("0");   //DIR = 0 for frm
	(rm).append("0");   //cell generated by a source
	(rm).append("0");   //CI bit by source
	(rm).append("0");   //NI initial
	(rm).append("x");   //Req/ACK unused
	(rm).append("xxx"); //reserved	


	//ECR
	char buf[16];
	sprintf(buf,"%d", reassembly->txtp->pcr);
	string s((16 - strlen(buf)),'0');
	(rm).append(s);
	(rm).append(buf);


	//CCR
	
	sprintf(buf,"%d",reassembly->txtp->acr);
	s = string((16 - strlen(buf)),'0');
	(rm).append(s);
	(rm).append(buf);


	//MCR

	sprintf(buf,"%d",reassembly->txtp->mcr);
	s = string((16 - strlen(buf)),'0');
	(rm).append(s);
	(rm).append(buf);

	// qlength - unused in ABR
	s = string(32,'x');
	(rm).append(s);

	//seq # unused in ABR
	(rm).append(s);

	//Reserved bits
	s = string(246,'x');
	(rm).append(s);

	//CRC
	s = string(10,'1');
	(rm).append(s);

	// assign payload

	//	printf("atmh->payload = 0x%x, length = %d\n",atmh->payload,(atmh->payload).length());

	atmh->payload = new string(rm);
	//atmh->payload.resize(atmh->payload.length() + rm.length());
	//(atmh->payload).assign(rm);

	return p;

}

void AAL5::adjust_source_rate (Packet *rm) {
	
	hdr_atm *atmh = hdr_atm::access(rm);
	
	string *s = atmh->payload;
	AtmTrafprm *t = reassembly->txtp;
	int ecr = atoi((atmh->payload->substr(16,16)).c_str());


	if ((s->at(10) == '0') && (s->at(11) == '0')) {
		double temp = min((double)t->pcr, t->acr + (t->rif * t->pcr));
		double temp1 = min((double)ecr,temp);
		t->acr = (int)(max((double)t->mcr, temp1));
	}

	else if ((s->at(10) == '1') && (s->at(11) == '0')) 
		t->acr = max(t->mcr, (int)min((double)ecr,t->acr * (1 - t->rdf)));

	else if ((s->at(10) == '0') && (s->at(11) == '1')) 
		t->acr = max(t->mcr, min(ecr,t->acr));

	else if ((s->at(10) == '1') && (s->at(11) == '1'))
		t->acr = max(t->mcr,(int)min((double)ecr,t->acr * (1 - t->rdf)));
}
	

int AAL5::allowable_cell_cnt() {

	Tcl &tcl = Tcl::instance();

	AtmTrafprm *t = reassembly->txtp;
	double pt = prev_time;
	double curr_time = Scheduler::instance().clock();

	if (prev_time > curr_time)
		tcl.resultf("Scheduler is running backwards in time\n");

	prev_time = curr_time;
	return (int)((curr_time - pt) * t->acr);
}

void AAL5::reverse_addresses_for_brm(Packet *p) {


	hdr_ip *iph = hdr_ip::access(p);
	hdr_ip *orig = iph;

        printf("AAL5::src = %d, dst = %d\n", iph->saddr(),iph->daddr());

        int temp_addr = orig->saddr();
        int temp_port = orig->sport();

	iph->saddr() = orig->daddr();
	iph->sport() = orig->dport();
	iph->daddr() = temp_addr; 
	iph->dport() = temp_port;

        printf("AAL5::new src = %d, new dst = %d\n", iph->saddr(),iph->daddr());
}


