/* -*- Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 2000 Regents of the University of New Hampshire.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the MASH Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /opt/src/mash/repository/vint/ns-2/cell.h,v 1.0 2000/07/22 22:46 sow Exp $
 *
 * Contributed by Sowmya Manjanatha, Dept. of Computer Science, UNH, July 2000 
 */

#ifndef __aal_h__
#define __aal_h__

#include "agent.h"
#include "packet.h"
#include "atm.h"
#include "atmqos.h"
#include <string>
#include <math.h>
#include <vector>
// #include "general.h"

#define SAMPLERATE 8000

class AAL5 : public TclObject {

public:
	AAL5() : aal_uid(0),prev_time(0.0) { reassembly = new Reassembly;}
	~AAL5() { delete reassembly; }
	Packet **segment(Packet *p);
	void initpkt(Packet *p);
	void set_parameters(Reassembly *r,string);
	// int is_member(int,int);
	void increment();
	void rx_increment();
        bool cellcount_matches();
	Reassembly get_reassembly() { return *reassembly; }
	Packet *build_frm(Packet *);                 // ABR only
	void adjust_source_rate(Packet *);   // ABR only
	vector<Packet *> get_abr_buffer() { return abr_buffer; }
	int allowable_cell_cnt();
	void store_in_abr_buffer(Packet *p) { abr_buffer.push_back(p); }
	void set_efci() { efci = true; }
	bool get_efci() { return efci; }
	void reverse_addresses_for_brm(Packet *);
	string aal_type() { return aal_type_;}
	int abr_buffer_size() { return abr_buffer.size(); }
	Packet *pop_abr() { 
		Packet *p;

		if (abr_buffer.size() > 0) {
			p = (Packet *)(*abr_buffer.begin());
			abr_buffer.erase(abr_buffer.begin());
			return p;
		}

		return (Packet *)NULL;
	}

protected:
	int command(int argc, const char*const* argv);
	
private:
	Reassembly *reassembly;
	double prev_time;
	vector<Packet *> abr_buffer;
	bool efci;
	int aal_uid;
	string aal_type_;

};


#endif



