/* -*- Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */

/*
 * Copyright (c) 1999 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the MASH Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /opt/src/mash/repository/vint/ns-2/atmnode.h,v 1.0 2000/06/19 17:35:09 sow Exp $
 *
 * Contributed by Sowmya Manjanatha, Dept. of Computer Science, UNH, August 2000 
 */

#include "atmconditioner.h"
#include "atmpvc.h"
#include "ip.h"

static class AtmConditionerClass : public TclClass {
public:
	AtmConditionerClass() : TclClass("AtmConditioner") {}
	TclObject* create(int, const char*const*) {
		return (new AtmConditioner);
		
	}	
} class_atmconditioner;


AtmConditioner::AtmConditioner():qos(0),bucketsize(50000),available_bucket(50000),buffer_limit(50000) 
{ 
	bucketlimit = 1000;
	
	tx_cells = 0;
	rx_cells = 0;
	drops = 0;
	non_cfm = 0;
}


int AtmConditioner::command(int argc, const char*const* argv) {


	if (argc == 2) {
	  if (strcmp (argv[1], "conditioner_statistics") == 0) {
		  printf("   %8d      %12d      %5d     %5d     \n",rx_cells,tx_cells,drops,non_cfm);
	  
		  return TCL_OK;
	  }
	}
		

	if (argc == 3) {
		
		if (strcmp("set-profile",argv[1]) == 0) {
			return TCL_OK;
		}
		
		if (strcmp("add-pvc_dbase",argv[1]) == 0) {
			pvc_dbase = (pvc_db *) TclObject::lookup(argv[2]);
			return TCL_OK;
		}
		
		if (strcmp("total_service_rate_", argv[1]) == 0) {
			total_service_rate = atof(argv[2]);
			available_service_rate = total_service_rate;
			printf("AtmConditioner::cac, avail service rate = %e\n", available_service_rate);
			
			if(total_service_rate < 0)
			  return TCL_ERROR;
			return TCL_OK;
		}
	}
	
	return (Connector::command(argc,argv));
}

/* This is where the cell first enters into the ATM Network.
   Well, this is because the traffic entering is conditioned
   upfront given to the fact that the AtmConditioner is installed
   between a link and a node. */


void AtmConditioner::recv(Packet *p, Handler *h) {

	int idx;

  /* Here I follow a strict allocation procedure. 
     If traffic is conforming, I allocate, else I drop */

	hdr_atm *atmh = hdr_atm::access(p);

        AtmQos *qos = classify(p);

	rx_cells++;

	if (qos == NULL) {
		printf("AtmConditioner :: qos is NULL, dropping packet\n");
		drops++;
		drop(p);
	}

	
	else
		{
			switch (qos->get_rxtp()->traffic_class) {
				
			case CBR:
			case rt_VBR:
			case nrt_VBR:
		//		store_cell(p);
				usage_parameter_control(p,h);   
				
				break;
			case GFR:
				gfr_upc(p,h);
			case ABR:
			case UBR:
				// traffic_shaping(p,h,(double)1/bucketrate);

				//				printf("AtmConditioner:: sending packet to target, vpi/vci = %d/%d, type = %d\n",atmh->vpi,atmh->vci,atmh->type);
				tx_cells++;
				target_->recv(p,h);
				break;
			}

		}
}

/* here again, I haven't worried about incoming PVC or rxtps.  
   I haven't the slightest idea where to find information re:
   their significance. */

void AtmConditioner::usage_parameter_control(Packet *p,Handler *h) {

	Tcl &tcl = Tcl::instance();
	double T;
	bool conforms;
        int type;

	AtmQos *qos = classify(p);

        if (qos == NULL) {
		drops++;
		drop(p);
		return;
         }

        type = qos->get_rxtp()->traffic_class;
	
	
	switch (type) {
		
	case CBR: 
		{
			int pcr = qos->get_rxtp()->pcr;
			T = (double) (1.0/(double)pcr);
			
			conforms = gcra(T,qos->get_rxtp()->cdvt, qos);
			upc_helper(conforms,p,h); 
			
		}
		break;
		
	case rt_VBR:
	case nrt_VBR:
		{
			double Ts = (double)(1.0/(double)(qos->get_rxtp()->scr));
			T = (double) (1.0/(double)(qos->get_rxtp()->pcr));
			double tau_s = qos->get_rxtp()->cdvt;
			int mbs = qos->get_rxtp()->mbs;
			
			double tau = (mbs - 1)*(Ts - T);
			conforms = gcra(Ts,(tau+tau_s),qos);
			upc_helper(conforms,p,h);
		}
		break;
	default: 
		tcl.resultf("invalid service category %d\n",type);
		break;
		
			
	}


}

bool AtmConditioner::gcra(double T, double tau, AtmQos *qos) {


	double tat = qos->get_rxtp()->tat;

	/* the first cell on the connection */
	if (tat == -1) {
		qos->get_rxtp()->tat = Scheduler::instance().clock() + T;
		return true;
	}
	
	double temp = Scheduler::instance().clock();


	/* conforming */
	if (temp >= tat) {
		qos->get_rxtp()->tat += T;
		return true;
	}

	else if ((temp < tat) && (tat <= (temp+tau))) {
		qos->get_rxtp()->tat = temp+T;
		return true;
	}

	else if (temp < (tat - tau)) {
                non_cfm++;
		return false;
        }

	return false;
}
		
		
		
	
/* iat stands for interarrival time */
// void AtmConditioner::traffic_shaping(Packet *p, Handler *h, double iat) {

//	update_tokens(iat);

	/* actual shaping function */

	/* if the bucketsize > 0 and buffer is not empty, 
           the bucket must have been updated in this run of
           traffic shaping function */

//	if (bucketsize > 0) {
		
/*		if (buffer.size() > 0) {
		
			do {
				bucketsize--;
				printf("AtmConditioner: %s sending packet to target\n",name());
				target_->recv(buffer.front(),h);
			} while ((buffer.size() > 0) && (bucketsize > 0));
		
			} */
		/* sufficient time to collect another token might
                   have elapsed after the while loop */
		
/*		if (bucketsize > 0) {
			printf("AtmConditioner: %s sending packet to target\n",name());
			target_->recv(p,h);
			bucketsize--;
		} else {
			if (buffer_limit >= buffer.size())
				{
					printf("AtmConditioner::dropping, time = %e\n", Scheduler::instance().clock());
					drop(p);
				}
			else
				buffer.push_back(p);
		}
		
	}
	
	else {
		
		if (buffer_limit >= buffer.size()) {
			printf("AtmConditioner::dropping packet, time = %e\n", Scheduler::instance().clock());
			drop(p);
		}
		else
			buffer.push_back(p);
	}

	} */
	 
/* connection admission control */
bool AtmConditioner::cac(path pth, int linknum) {

	AtmTrafprm *txtp = pth.outgoing_pvc.get_txtp();

	// transmit side parameter check
	
	Tcl &tcl = Tcl::instance();

	switch (txtp->traffic_class) {

	case CBR:
		/* for now, I am just considering conditioning on
                   the outgoing PVC and also only the transmit
                   traffic descriptor information.  I will have
                   to find out the significance of the rx traffic
                   descriptor as well as the incoming pvc */

		if ((txtp->pcr * CELLSIZE * BYTE) >= pvc_dbase->avail_rate(linknum)) {
			return false;
		}
		else {
			pvc_dbase->set_avail_rate((txtp->pcr) * CELLSIZE * BYTE,linknum);
			return true;
		}
		
		break;

	case rt_VBR:


		if ((CELLSIZE * BYTE *((1+A1)*txtp->scr)) >= pvc_dbase->avail_rate(linknum)) {

			printf("AtmCond::cac, requested rate = %e,scr = %d\n",CELLSIZE*BYTE*((1+A1)*txtp->scr),txtp->scr);
			return false;
		}
		else {
			pvc_dbase->set_avail_rate( (txtp->scr) * CELLSIZE * BYTE,linknum);
			return true;
		}
		break;

	case nrt_VBR:

		if ((txtp->scr * CELLSIZE * BYTE )>= pvc_dbase->avail_rate(linknum))
			return false;
		else {
			pvc_dbase->set_avail_rate((txtp->scr) * CELLSIZE * BYTE,linknum);
			return true;
		}

		break;

	case ABR:
	case GFR:

		if ((txtp->mcr * CELLSIZE * BYTE) >= pvc_dbase->avail_rate(linknum))
			return false;
		else {
			pvc_dbase->set_avail_rate((txtp->mcr) * CELLSIZE * BYTE,linknum);
			return true;
		}
		break;

	case UBR:
		break;

	default:
		tcl.resultf("unknown ATM traffic class");
	}
	return true;
}

AtmQos *AtmConditioner::classify(Packet *p) {

	hdr_atm *atmh = hdr_atm::access(p);

	return pvc_dbase->incoming_qos_lookup(atmh->vpi,atmh->vci);
}




void AtmConditioner::upc_helper_old(bool conforms, Packet *p, Handler *h) 
{
	hdr_atm *atmh = hdr_atm::access(p);
	Tcl &tcl = Tcl::instance();

	int idx = find_bucket(p);

	if (conforms && ((atmh->clp == 0) || (atmh->clp == 1))) 
		{
			
			if (idx  < 0) {
				tcl.resultf("AtmConditioner::traffic unprovisioned");
				return;
			}
				

			if (leakybucket[idx]->size() > 0) {
				Packet *temp = leakybucket[idx]->first();
				leakybucket[idx]->erase();
				tx_cells++;
				target_->recv(temp,h);
			}
			// traffic_shaping(leakybucket[idx]->first(),h,(double)1/bucketrate);

		}
	
	else if (!conforms && (atmh->clp == 0)) {
		atmh->clp = 1;
	}
	
	else if (!conforms && (atmh->clp == 1)) {
		//	printf("AtmConditioner :: ID %s dropping packet at %e\n",name(),Scheduler::instance().clock());
		drops++;
		drop(leakybucket[idx]->first());
		leakybucket[idx]->erase();
	}
	
}


/* since NS is a single thread simulator, tokens into a bucket cannot
   be generated parallely without being interrupted by other tasks.
   Therefore, I use an approximation to the token generation concept.

   I maintain a time variable that is basically the time at which the
   previous token was generated.  The floor or integer value of
   elapsed_time/token_bucket_rate gives the number of tokens that must
   be input into the bucket. */

 void AtmConditioner::update_tokens (double iat) {
	
	 double temp = Scheduler::instance().clock();
	 double elapsed_time = temp - last_token_update_time;
	 last_token_update_time = temp;
	 
	 bucketsize += (int)(elapsed_time/iat);
 }


 int AtmConditioner::find_bucket(Packet *p)
	 {

		 hdr_atm *atmh = hdr_atm::access(p);

		 for (unsigned int i = 0; i < leakybucket.size(); i++)
			 {
				 int vp = (int)leakybucket[i]->vpi();
				 int vc = (int)leakybucket[i]->vci();
				 
				 if ((vp == atmh->vpi) && (vc == atmh->vci))
					 return i;
			 }

		 return -1;
	 }
				 
 void AtmConditioner::store_cell(Packet *p)
	 {

		 int idx = find_bucket(p);
		 hdr_atm *atmh = hdr_atm::access(p);

		 if (idx > -1) {
			 if (leakybucket[idx]->size() < bucketlimit)
				 leakybucket[idx]->push_back(p);
			 
			 else {
				 drops++;
				 drop(p);
			 }
		 }
		 else {
			 VCBuffer *l = new VCBuffer(atmh->vpi,atmh->vci);
			 leakybucket.push_back(l);
			 leakybucket[leakybucket.size() - 1]->push_back(p);

		 }
		 
		 return;
	 }

char *decode_type(int type)
{
	if (type == CBR)
		return "CBR";
	
	else if (type == rt_VBR)
		return "rt-VBR";

	else if (type == nrt_VBR)
		return "nrt-VBR";

	else if (type == GFR)
		return "GFR";

	else if (type == ABR)
		return "ABR";

	else if (type == UBR)
		return "UBR";

	else
		return "UNKNOWN";
}

// the policing for gfr is frame based and hence the calculations 
// vary from normal gcra.

void AtmConditioner::gfr_upc(Packet *p, Handler *h) {

	bool clp  = gfr_clp(p);
	bool gcra = f_gcra(p);
	bool mfs  = gfr_mfs(p);

	if (gcra && mfs && clp) {
		tx_cells++;
		target_->recv(p,h);
	}

	else {
		drops++;
		drop(p);
	}

	return;
}


// frame based gcra

bool AtmConditioner::f_gcra(Packet *p) {

	hdr_atm *atmh = hdr_atm::access(p);
	
	AtmQos *qos = classify(p);
        AtmTrafprm *rxtp  = qos->get_rxtp();

	double T = 1/(rxtp->mcr);
	double BT = ((rxtp->mbs) - 1)*((1/rxtp->mcr) - (1/rxtp->pcr));
	double L = BT + rxtp->cdvt;

	printf("%e  %e    %e 0x%x\n",T,BT,L,atmh);


	return gcra(T,BT+rxtp->cdvt,qos);
}


bool AtmConditioner::gfr_mfs(Packet *p) {

	hdr_atm *atmh = hdr_atm::access(p);
	AtmTrafprm *rxtp = classify(p)->get_rxtp();
	int gfr_idx = find_gfr_cell_cnt(atmh->vpi, atmh->vci);
	int gcnt = gfr_cells_per_vc[gcnt].gfr_cells;

	if (( ((atmh->type%2) != 1) && (rxtp->mfs > gcnt)) || (atmh->type%2 == 1))
		{
			gfr_cells_per_vc[gfr_idx].gfr_cells++;
			return true;
		}

	return false;
}


bool AtmConditioner::gfr_clp(Packet *p) {

	hdr_atm *atmh = hdr_atm::access(p);

	int idx = find_gfr_cell_cnt(atmh->vpi,atmh->vci);

	// this is the first cell on the vc
	if (idx == -1) {
		gfr_info info = gfr_info(atmh->vpi, atmh->vci, atmh->clp);
		gfr_cells_per_vc.push_back(info);
	}


	else {
		if (atmh->clp == gfr_cells_per_vc[idx].clp)
			return true;
		return false;
	}
		
	return false;
}


int AtmConditioner::find_gfr_cell_cnt(int vpi, int vci) {

	for(int i=0; i < gfr_cells_per_vc.size() ; i++) {
		int vp = gfr_cells_per_vc[i].vpi;
		int vc = gfr_cells_per_vc[i].vci;

		if ((vp == vpi) && (vc == vci))
			return i;
	}

	return -1;
}


void AtmConditioner::upc_helper(bool conforms,Packet *p, Handler *h) {

         int idx = find_bucket(p);
         hdr_atm *atmh = hdr_atm::access(p);


         if (conforms) {
                if (idx < 0) {
                      tx_cells++;
                      target_->recv(p,h);
                }


                else {
                    Packet *c = (Packet *)leakybucket[idx]->first();
                    leakybucket[idx]->erase();
                    store_cell(p);
                    tx_cells++;
                    target_->recv(c,h);
                }
           }

           else {

                store_cell(p);
             }


}


