
/* -*- Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */

/*
 * Copyright (c) 1999 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the MASH Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /opt/src/mash/repository/vint/ns-2/atmnode.h,v 1.0 2000/08/09 20:43:00 sow Exp $
 *
 * Contributed by Sowmya Manjanatha, Dept. of Computer Science, UNH, August 2000
 */

#ifndef __atmconditioner_h__
#define __atmconditioner_h__

#include "connector.h"
#include "packet.h"
#include "atm.h"
#include "atmqos.h"
#include <vector>
#include "leakybucket.h"

typedef struct path path;
class pvc_db;

// connection parameter constants for CAC

#define A1 0
#define A2 0
#define A3 0
#define A4 0
#define A5 0
#define A6 0
#define A7 0
#define A8 0
#define A9 200

class AtmConditioner: public Connector {
public:
  AtmConditioner();
  void recv(Packet *p, Handler *h);
  bool cac(path,int);
  
protected:
  int command(int argc, const char*const* argv);	
  void usage_parameter_control(Packet *, Handler *);
  bool gcra(double, double, AtmQos *);
  void traffic_shaping(Packet *, Handler *,double);
  AtmQos *classify(Packet *p);
  void upc_helper(bool, Packet *, Handler *);
  void upc_helper_old(bool, Packet *, Handler *);
  void update_tokens(double);
  int find_bucket(Packet *);
  void store_cell(Packet *);
  void gfr_upc(Packet *, Handler *);
  bool f_gcra(Packet *);
  bool gfr_mfs(Packet *);
  bool gfr_clp(Packet *);
  int find_gfr_cell_cnt(int,int);
  
private:
  int qos;
  int bucketsize;
  double bucketrate;
  int available_bucket;
  pvc_db *pvc_dbase;
  double total_service_rate;
  double available_service_rate;
  vector <Packet *> buffer;
  unsigned int buffer_limit;
  double last_token_update_time;
  vector <VCBuffer *> leakybucket;
  int bucketlimit;
  vector <gfr_info> gfr_cells_per_vc;

  //statistics

  unsigned int rx_cells;
  unsigned int tx_cells;
  unsigned int drops;
  unsigned int non_cfm;

};

char *decode_type(int);



#endif



