/* -*- Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 2000 Regents of the University of New Hampshire.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the MASH Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /opt/src/mash/repository/vint/ns-2/atmnode.h,v 1.0 2000/07/20 14:50 sow Exp $
 *
 * Contributed by Sowmya Manjanatha, Dept. of Computer Science, UNH, July 2000 
 */

#ifndef __atmqueue_h__
#define __atmqueue_h__

#include "packet.h"
#include <vector>

#define CBR     0
#define rt_VBR  1
#define nrt_VBR 2
#define ABR     3
#define UBR     4
#define GFR     5

#define Q_DEF   200

class AtmQueue {
 public:
	AtmQueue() : qlength(2000) { }
	AtmQueue(int vp, int vc) : qlength(Q_DEF),vpi(vp),vci(vc),q_in(0),q_drop(0) { }
	Packet *deque(int clp) {
		if (queue[clp].size() > 0) {
			Packet *temp = (Packet *)(*(queue[clp].begin()));
			queue[clp].erase(queue[clp].begin());
			return temp;
		}

		return (Packet*)NULL;
	}
	int get_vpi() { return vpi;}
	int get_vci() { return vci;}
        void set_qlength(int qlen) { qlength = qlen; }
	void set_vpi(int vp) { vpi = vp; }
	void set_vci(int vc) { vci = vc; }
	vector<Packet *> get_queue_clp0() { return queue[0]; }
	vector<Packet *> get_queue_clp1() { return queue[1]; }
		
	bool enque(Packet *p,int clp) { 
		q_in++;
		if ((int)queue[clp].size() <= qlength) {
			queue[clp].push_back(p);
			return true;
		}
		// printf("AtmQueue::enque, qlength exceeded\n");
		q_drop++;
		return false;
	}

	unsigned int qin_cnt () { return q_in; }
	unsigned int qdrop_cnt () { return q_drop; }
		
  
private:
	vector<Packet *> queue[2];       /* the six queues are representative of six
                                            service categories in atm */
	int qlength;
	int vpi;
	int vci;
	unsigned int q_in;
	unsigned int q_drop;
	
};

#endif



