/* -*- Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 2000 Regents of the University of New Hampshire.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the MASH Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /opt/src/mash/repository/vint/ns-2/atmscheduler.h,v 1.0 2000/07/22 22:46 sow Exp $
 *
 * Contributed by Sowmya Manjanatha, Dept. of Computer Science, UNH, July 2000 
 */

#include "atmscheduler.h"

static class AtmSchedulerClass : public TclClass {
public:
	AtmSchedulerClass() : TclClass("AtmScheduler") {}
	TclObject* create(int, const char*const*) {
		return (new AtmScheduler);
		
	}	
} class_atmscheduler;


inline void AtmSchedulerHandler::handle(Event *e)
{
	scheduler->resume();
}

AtmScheduler::AtmScheduler() : Connector(),handler(this),idle_(true) { 

	last_served_category = -1;
	
	for (int i=0; i < NUM_CATEGORIES; i++)
		last_served_queue_in_category[i] = -1;

	for (int i=0; i < MAX_Q; i++)
		cells_per_round[i] = -1;

	bind ("bytes-per-round",&bytes_per_round);

	tx_cells = 0;
	rx_cells = 0;
	drops = 0;
	undequeued  = 0;
        
        for (int i = 0; i < NUM_CATEGORIES; i++) {
            for (int j=0; j < MAX_Q; j++) {
                qWeight[i][j] = 1;
                wrrTemp[i][j] = 0;
            }
        }

        last_served_q = -1;
}


void AtmScheduler::recv(Packet *p, Handler *h) {

	hdr_atm *atmh = hdr_atm::access(p);

	update_total_input_rate(atmh->vpi,atmh->vci);

	rx_cells++;


	bool enqueued = enque(p);

	if (enqueued == false) 
		{
			drops++;
			drop(p);
		}
	
	
	if (idle_ && enqueued) {
		idle_ = false;
		assign_service_sizes_per_category();
		resume();
	}  
	

}

void AtmScheduler::resume() {

	Packet *c;


	c = deque();

	if (c == NULL) {
		undequeued++;
		// printf("AtmScheduler::atm_queue length is zero\n");
		idle_ = true;
		return;
	}

	tx_cells++;
	target_->recv(c,&handler);  
	
	
}

int AtmScheduler::command(int argc, const char*const* argv) {

	unsigned int temp = 0;
	
	if (argc == 2) {
	  if (strcmp (argv[1], "scheduler_statistics") == 0) {
		  for (int i=0; i < (int)per_vc_q[UBR].size() ; i++) 
			  temp += per_vc_q[UBR][i].qin_cnt();

		  printf("   %8d     %12d       %7d   %5d   %5d   \n",rx_cells,tx_cells,drops,temp,undequeued);
	  
		  return TCL_OK;
	  }
	}
	


	if (argc == 3) {
		// assigning queue lengths
		if (strcmp("queue-length-cbr",argv[1]) == 0) {
			cbr_qlength = atoi(argv[2]);

			if (cbr_qlength < 0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-length-rt-vbr",argv[1]) == 0) {
			rt_vbr_qlength = atoi(argv[2]);

			if (rt_vbr_qlength < 0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-length-nrt-vbr",argv[1]) == 0) {
			nrt_vbr_qlength = atoi(argv[2]);

			if (nrt_vbr_qlength < 0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-length-abr",argv[1]) == 0) {
			abr_qlength = atoi(argv[2]);

			if (abr_qlength < 0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-length-ubr",argv[1]) == 0) {
			ubr_qlength = atoi(argv[2]);

			if (ubr_qlength < 0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-length-gfr",argv[1]) == 0) {
			gfr_qlength = atoi(argv[2]);

			if (gfr_qlength < 0)
				return TCL_ERROR;
			return TCL_OK;
		}

		//assigning queue weights
		if (strcmp("queue-weight-cbr",argv[1]) == 0) {
			cbr_qweight = atof(argv[2]);

			if (cbr_qlength < 0.0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-weight-rt-vbr",argv[1]) == 0) {
			rt_vbr_qweight = atof(argv[2]);

			if (rt_vbr_qweight < 0.0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-weight-nrt-vbr",argv[1]) == 0) {
			nrt_vbr_qweight = atof(argv[2]);

			if (nrt_vbr_qweight < 0.0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-weight-abr",argv[1]) == 0) {
			abr_qweight = atof(argv[2]);

			if (abr_qweight < 0.0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-weight-ubr",argv[1]) == 0) {
			ubr_qweight = atof(argv[2]);

			if (ubr_qweight < 0.0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("queue-weight-gfr",argv[1]) == 0) {
			gfr_qweight = atof(argv[2]);

			if (gfr_qweight < 0.0)
				return TCL_ERROR;
			return TCL_OK;
		}

		if (strcmp("attach-pvc_dbase",argv[1]) == 0) {
			pvc_dbase = (pvc_db *) TclObject::lookup(argv[2]);
			return TCL_OK;
		}
	}
        if (argc == 5) {
                if (strcmp("set-qlength",argv[1]) == 0) {
                        printf("set-qlength\n");
                        return set_qlength(argv);
                }
                if (strcmp("set-qweight",argv[1]) == 0) {
                        printf("set-qweight\n");
                        return set_qweight(argv);
                }
        }
             

	return (Connector::command(argc,argv));
}


int AtmScheduler::set_qlength(const char*const* argv) {

    int qnum = atoi(argv[3]);
    int qlen = atoi(argv[4]);

    printf("qlength  = %d,qnum = %d\n",qlen,qnum);

    for (int i = 0; i < NUM_CATEGORIES ; i++) {
          if (strcmp(argv[2], qosT[i]) == 0) {
                qlen_[i][qnum] = qlen;
                return TCL_OK;
          }
    }
    return TCL_ERROR;
}


int AtmScheduler::set_qweight(const char*const* argv) {

    int qnum = atoi(argv[3]);
    int qwt = atoi(argv[4]);

    printf("set qwt. qweight  = %d,qnum = %d\n",qwt,qnum);

    for (int i = 0; i < NUM_CATEGORIES ; i++) {
          if (strcmp(argv[2], qosT[i]) == 0) {
                qWeight[i][qnum] = qwt;
                return TCL_OK;
          }
    }
    return TCL_ERROR;
}

void AtmScheduler::reset() {

}


//finds the per-vc-queue given vpi and vci values
 int AtmScheduler::pvc_found(int vp, int vc,int type) {
	
	 for (unsigned int i=0; i < per_vc_q[type].size(); i++) {
		 if ((vp == per_vc_q[type][i].get_vpi()) && (vc == per_vc_q[type][i].get_vci()))
			 return (int)i;
	 }
	 
	 return -1;
 }
 
 bool AtmScheduler::enque(Packet *p) {
	 hdr_atm *atmh = hdr_atm::access(p);
	 AtmTrafprm *txtp;

	 PVC *pvc = pvc_dbase->get_corresponding_outgoing_pvc(atmh->vpi,atmh->vci);


	 if (pvc != NULL)
		 txtp = pvc->get_txtp();
	 
	 else {
		 printf("AtmScheduler::packet vpi/vci = %d/%d\n",atmh->vpi,atmh->vci);

		 printf("AtmScheduler::enque, PVC NOT found\n");
		 return false;
	 }
	 int i;
	 
	 if ((i=pvc_found(atmh->vpi,atmh->vci,txtp->traffic_class)) > -1) {
		 if (atmh->clp == 0)
			 return per_vc_q[txtp->traffic_class][i].enque(p,0);
		 else 
			 return per_vc_q[txtp->traffic_class][i].enque(p,1);
	 }
	 
	 else {
		 AtmQueue temp_q(atmh->vpi,atmh->vci);
                 int cl = txtp->traffic_class;
                 temp_q.set_qlength(qlen_[cl][per_vc_q[cl].size()]);
                 printf("AtmScheduler:: cl = %d, q = %d,size = %d\n",cl,per_vc_q[cl].size(),qlen_[cl][per_vc_q[cl].size()]);
		 per_vc_q[txtp->traffic_class].push_back(temp_q);
		 i = per_vc_q[txtp->traffic_class].size() - 1;
		 if (atmh->clp == 0)
			 return per_vc_q[txtp->traffic_class][i].enque(p,0);
		 else 
			 return per_vc_q[txtp->traffic_class][i].enque(p,1);
	 } 
 }
 
 
 void AtmScheduler::assign_service_sizes_per_category () {
	 
	 double b = (double)bytes_per_round;
	 double c = (double)CELLSIZE;

	 cells_per_category[CBR]     = (int)((b * cbr_qweight)/c);
	 cells_per_category[rt_VBR]  = (int)((b * rt_vbr_qweight)/c);
	 cells_per_category[nrt_VBR] = (int)((b * nrt_vbr_qweight)/c);
	 cells_per_category[ABR]     = (int)((b * abr_qweight)/c);
	 cells_per_category[UBR]     = (int)((b * ubr_qweight)/c);
	 cells_per_category[GFR]     = (int)((b * gfr_qweight)/c);
 }
 

void AtmScheduler::update_total_input_rate(int vp, int vc) {

	path *entry = pvc_dbase->get_corresponding_table_entry(vp,vc);
	double t =  Scheduler::instance().clock();

	if (entry != NULL) {
		double curr_rate = 1/(t - entry->prev_cell_arrival_time);
		entry->input_rate = (entry->input_rate + curr_rate)/2;
		total_input_rate = pvc_dbase->get_total_input_rate();
	}

}
	
Packet *AtmScheduler::deque_old() {
	
	Packet *p;
	int q_trav=0;
	
	
	int curr_ctgry = last_served_category+1;
	int curr_q     = last_served_queue_in_category[curr_ctgry]+1;
	int ctgries    = NUM_CATEGORIES;
	int ctgry_trav=0;
	
	
	int i = curr_ctgry;
	while (i < ctgries)
		{
			ctgry_trav++;
			p = per_vc_queuing(curr_q,per_vc_q[i].size(),i,q_trav);
			
			if(p)
				{
					last_served_category=i;
					return p;
				}

			else if((!p) && (q_trav < (int)per_vc_q[i].size())) 
				{
				if(per_vc_queuing(0,curr_q,i,q_trav) != NULL)
					{
						last_served_category = i;
						return p;
					}
				}

			if ((i==(ctgries-1)) && (ctgry_trav < ctgries))
				{
					i=-1;
					ctgries=curr_ctgry;
				}

			i++;
			
		}
	
	idle_ = true;
	return (Packet *)NULL;
}

/* Description :  per_vc_queueing traverses the list of queues w.r.t the category input
                  and finds a packet if available on a fifo basis */
 
Packet *AtmScheduler::per_vc_queuing(int start, int end, int ctgry, int &q_trvsed)
{
	Packet *p;

	for(int i=start; i < end; i++) 
		{
			for(int clp=0; clp < 2; clp++)
				{
					q_trvsed++;
					if((p=per_vc_q[ctgry][i].deque(clp)) != NULL) {
							last_served_queue_in_category[ctgry] = i;
							return p;
					}
				}  
					    
		}

	return (Packet *)NULL;

}


Packet *AtmScheduler::deque() {

	
//         return priority_scheduler();
	 return wrr_scheduler();
}
	


/* This is a simple priority scheduler, this method may be used both for per_vc_q scheduling and
   per_category scheduling where CBR gets the highest priority and UBR gets the least */

Packet *AtmScheduler::priority_scheduler() {
	
	Packet *p;

	for (int i=0; i < NUM_CATEGORIES ; i++)
		if((p=pri_vc_scheduler(i)) != NULL)
			return p;

	return (Packet*)NULL;
}

Packet *AtmScheduler::pri_vc_scheduler(int ctgry) {
	
	Packet *p;

	for (int i=0 ; i < (int)per_vc_q[ctgry].size() ; i++) {

		for (int clp=0; clp < 2; clp++)
			if ((p=per_vc_q[ctgry][i].deque(clp)) != NULL)
				return p;
	}

	return (Packet *)NULL;
}
	
/* ********************************************************* */
/*  This is a rate pacing scheduler.  The rate pacing happens 
    according to the weights for each PVC which further depends
    on individual parameters such as PCR for CBR, SCR for VBR
    and MCR for ABR */
/* ********************************************************* */
	
Packet *AtmScheduler::WRR_vc_scheduler(int ctgry) {

	Packet *p;

        int i = 0;
      
        /* *************************** */
        // to avoid floation point errors
        /* *************************** */

        if (per_vc_q[ctgry].size() == 0)
             return (Packet *)NULL;

        if(wrrTemp[ctgry][last_served_q] <= 0) {
             last_served_q = (last_served_q + 1) % per_vc_q[ctgry].size();
	     wrrTemp[ctgry][last_served_q] = qWeight[ctgry][last_served_q] - 1;
        } else {
             wrrTemp[ctgry][last_served_q] = wrrTemp[ctgry][last_served_q] - 1;
        }

        while (i < (int)per_vc_q[ctgry].size()) {
                
             for (int clp = 0; clp < 2; clp++) {
                 if((p=per_vc_q[ctgry][last_served_q].deque(clp)) != NULL) {
                     return p;
                }
              }

              wrrTemp[ctgry][last_served_q] = 0;
              last_served_q = (last_served_q + 1) % per_vc_q[ctgry].size();
              wrrTemp[ctgry][last_served_q] = qWeight[ctgry][last_served_q] - 1;
              i++;

        }
        
        return (Packet *)NULL;
}


Packet *AtmScheduler::wrr_scheduler() {
	
	Packet *p;

	for (int i=0; i < (NUM_CATEGORIES-1) ; i++)
		if((p=WRR_vc_scheduler(i)) != NULL)
			return p;

        /* ************************************ */
        /* GFR gets the least priority          */
        /* ************************************ */

        if ((p=wrr_vc_scheduler(NUM_CATEGORIES-1)) != NULL)
               return p;

        return (Packet *)NULL;
}


Packet *AtmScheduler::wrr_vc_scheduler(int ctgry) {
	
	Packet *p;

        int last = last_served_q;
        int i = 0;


        while (i < (int)per_vc_q[ctgry].size()) {

              last = (last + 1) % per_vc_q[ctgry].size();
                
              for (int clp = 0; clp < 2; clp++) {
                 if((p=per_vc_q[ctgry][last].deque(clp)) != NULL) {
                      last_served_q = last;
                      return p;
                 }
              }

              i++;

        }
        
        return (Packet *)NULL;
}



void AtmScheduler::cells_per_q(int ctgry) {

         for (int i = 0; i < (int)per_vc_q[ctgry].size(); i++) 
             cells_per_round[i] = CELLS_PER_ROUND;
}

             
/* ********************************************************* */
/*  GCRA algorithm used to decide on the rate pacing scheme
    mentioned in the WRR_vc_scheduler
    and MCR for ABR */
/* ********************************************************* */


bool AtmScheduler::gcra_helper(Packet *p,int qnum) {

	Tcl &tcl = Tcl::instance();
	double T;
	bool conforms;
        int type;
        hdr_atm *atmh = hdr_atm::access(p);

	AtmQos *qos = pvc_dbase->outgoing_qos_lookup(atmh->vpi,atmh->vci);

        if (qos == NULL) 
           return false;

        type = qos->get_rxtp()->traffic_class;
	
	switch (type) {
		
	case CBR: 
		{
			int pcr = qos->get_rxtp()->pcr;
			T = (double) (1.0/(double)pcr);
			
			conforms = gcra(T,qos->get_rxtp()->cdvt, type,qnum);
                        return conforms;
			
		}
		break;
		
	case rt_VBR:
	case nrt_VBR:
		{
			double Ts = (double)(1.0/(double)(qos->get_rxtp()->scr));
			T = (double) (1.0/(double)(qos->get_rxtp()->pcr));
			double tau_s = qos->get_rxtp()->cdvt;
			int mbs = qos->get_rxtp()->mbs;
			
			double tau = (mbs - 1)*(Ts - T);
			conforms = gcra(Ts,(tau+tau_s),type,qnum);
                        return conforms;
		}
		break;
	case ABR: 
		{
			int mcr = qos->get_rxtp()->mcr;
			T = (double) (1.0/(double)mcr);
			conforms = gcra(T,qos->get_rxtp()->cdvt,type,qnum );
                        return conforms;
		}
		break;
	default: 
		tcl.resultf("invalid service category %d\n",type);
		break;
		
			
	}
return false;

}

/* ********************************************************* */
/*  GCRA algorithm used to decide on the rate pacing scheme
    mentioned in the WRR_vc_scheduler
    and MCR for ABR */
/* ********************************************************* */

bool AtmScheduler::gcra(double T, double tau, int c, int q) {

	double tat = tat_[c][q];

	/* the first cell on the connection */
	if (tat == -1) {
		tat_[c][q] = Scheduler::instance().clock() + T;
		return true;
	}
	
	double temp = Scheduler::instance().clock();


	/* conforming */
	if (temp >= tat) {
		tat_[c][q] += T;
		return true;
	}

	else if ((temp < tat) && (tat <= (temp+tau))) {
		tat_[c][q] = temp+T;
		return true;
	}

	else if (temp < (tat - tau)) {
                //non_cfm++;
		return false;
        }

	return false;
}
