/* -*- Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 2000 Regents of the University of New Hampshire.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the MASH Research
 *      Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * @(#) $Header: /opt/src/mash/repository/vint/ns-2/sar.cc,v 1.0 2000/08/15 10:37 sow Exp $
 *
 * Contributed by Sowmya Manjanatha, Dept. of Computer Science, UNH, August 2000 
 */

#include <stdlib.h>
#include <string>
#include "config.h"
#include "sar.h"

char *decode_bool(bool);

static class ATMHeaderClass : public PacketHeaderClass {
public:
	ATMHeaderClass() : PacketHeaderClass("PacketHeader/ATM",
		sizeof(hdr_atm)) {}
} class_atmhdr;


static class SARDriverClass : public TclClass {
public:
  SARDriverClass() : TclClass("Agent/SARDriver") {}
  TclObject* create(int, const char*const*) {
    return (new SARDriver());
  }
} class_sardriver;


void SARTimer::expire(Event *)
{
	sar_->timeout();
}


SARDriver::SARDriver() : Agent(PT_ATM), timer_(this)
{
	bind("packetSize_", &size_);
	bind("delay_",&delay_);
	bind("bw_",&bw_);
	atm = false;
	atmswitch = false;
	cell_cnt_for_abr = -1;
	bw_ = 0.0;
	delay_ = 0.0;
	sar_q = new PacketQueue;
	alwd_cell_cnt = 0;
	aal = new AAL5;

	//initializing statistics variables
	rx_pkts = 0;
	tx_cells = 0;
	cells_per_pkt = 0;

}

SARDriver::SARDriver(packet_t type) : Agent(type), timer_(this)
{

}

void SARDriver::start_timer()
{
	timer_.sched(txtime()+delay_);

}

void SARDriver::timeout()
{

	Packet *p = sar_q->deque();


	if (p != NULL) {
                // printf("SARDriver::sar_q len = %d\n",sar_q->length());
		tx_cells++;
		target_->recv(p);
	}

	double t = next_time();

	if (t > 0)
		timer_.resched(t);

}


double SARDriver::next_time()
{
	return (txtime() + delay_);
}

	
int SARDriver::command(int argc, const char*const* argv)
{
  Tcl &tcl = Tcl::instance();

  if(argc == 2) {
	  if (strcmp(argv[1], "is_atmswitch") == 0) {
		  atmswitch = true;
		  return (TCL_OK);
	  }
	  
	  if (strcmp (argv[1], "set-atm") == 0) {
		  atm = true;
		  return (TCL_OK);
	  }

	  if (strcmp (argv[1], "sardriver_statistics") == 0) {
		  printf("   %8d             %12d                      \n",rx_pkts,tx_cells);
	  
		  return TCL_OK;
	  }
  }
  else if(argc == 3) {
	  if(strcmp(argv[1], "configure_atmswitch") == 0) {
		  pvc_database = (pvc_db*)TclObject::lookup(argv[2]);
		  if(pvc_database == NULL) {
			  tcl.resultf("SARDriver::invalid pvc_database value %s\n",argv[2]);
			  return TCL_ERROR;
		  }
		  return TCL_OK;
		  
	  }
	  
  } else if (argc == 4) {
	  
	  if (strcmp(argv[1], "start_timer") == 0) {
		  delay_ = atof(argv[2]);
		  bw_ = atof(argv[3]);
		  
		  start_timer();
	  }
	  
	  return TCL_OK;
	  
  } else if (argc == 7)  {   /* ATM extn - CNRG, UNH */
				// enabling IP over ATM support
	  if(strcmp(argv[1], "attach_aal5") == 0) {
		  
		  atm = true;
		  Reassembly *r= new Reassembly;
		  r->vpi = atoi(argv[2]);
		  r->vci = atoi(argv[3]);
		  r->rxtp = (AtmTrafprm*)TclObject::lookup(argv[4]);
		  r->txtp = (AtmTrafprm*)TclObject::lookup(argv[5]);
		  string type = string(argv[6]);

		  aal->set_parameters(r,type);
		  
		  
		  /* if abr, RM cells must be generated after every
		     cell_cnt_for_abr = (nrm - 1), hence initialize
		     cell_cnt_for_abr to 0.  Note this also indicates
		     whether traffic class is abr or not. */
		  
		  if((r->txtp)->traffic_class == ABR) {
			  cell_cnt_for_abr = 0;
		  }
		  
		  return TCL_OK;
	  }
    
    
  }
  
  return Agent::command(argc,argv);
}

void SARDriver::recv(Packet *p, Handler *h)
{

	string s = string("src_aal");
	string d = string("dst_aal");

        hdr_cmn *th = hdr_cmn::access(p);
        hdr_atm *atmh = hdr_atm::access(p);
 
	if ((atm) && !(p->is_atm()) && (!atmswitch) && (aal->aal_type() == s)) {
		rx_pkts++;
		aal5_segmentation(p,h);
	}

	else if ((atm) && (p->is_atm()) && (!atmswitch)) {
		printf("SAR::received brm from classifier, I am %s\n",name());
		send_brm(p,h);
	}

	else {
		send(p,h);
		//target_->recv(p,h);    
	}
}


void SARDriver::send_brm(Packet *p, Handler *h) {
	
	hdr_atm *atmh = hdr_atm::access(p);

	printf("SARDRIVER :: send_brm function, p = 0x%x\n",p);
	//this is a brm cell which must be sent back to source
	if ( (atmh->type == ATM_RM) && (atmh->payload->at(8) == '1')) {
		aal->reverse_addresses_for_brm(p);
		send(p,h);
	}

	else
		target_->recv(p,h);
}

// aal5_segmentation converts the incoming packets into ATM celss
// before forwarding it to the appropriate destination.

void SARDriver::aal5_segmentation(Packet *p, Handler *h) {

	Packet **c;
	int i = 0;
	Scheduler &s = Scheduler::instance();
	
	hdr_ip *iph = hdr_ip::access(p);
	c=aal->segment(p);


	AtmTrafprm *txtp = (aal->get_reassembly()).txtp;

	while (hdr_atm::access(c[i])->type == 0) {

		if (txtp->traffic_class == ABR) {

			process_abr(c[i],h);
		}

		else {
			sar_q->enque(c[i]);
		}

		i++;
		
	}

	// this is the last cell of the pdu

	if (txtp->traffic_class == ABR)
		process_abr(c[i],h);

	else 
		sar_q->enque(c[i]);

}

void SARDriver::resume()
{
	Packet *p = sar_q->deque();
	

}



/* If the source is of ABR type, then it is expected to adjust
   its rate dynamically with respect to the feedback from the 
   network.  This is done through a set of Resource Management (RM)
   cells in ATM.  The source generates an RM for every N data cells
   generated.  This part of the code is  taking care of this process.
*/


void SARDriver::process_abr(Packet *c, Handler *h) 
{
	alwd_cell_cnt += aal->allowable_cell_cnt();

	/* first send all the previously buffered cells */

	while ((alwd_cell_cnt > 0) && (aal->abr_buffer_size() > 0)) {
		Packet *ap = aal->pop_abr();
		printf("getting packet out of abr buffer,ap = 0x%x\n",ap);
		forward_abr(ap);
	}

	// now the received cell if allowable
	if (alwd_cell_cnt > 0) {
		forward_abr(c);
	}
	else {
		printf("SARDriver::storing in aal buffer\n");
		aal->store_in_abr_buffer(c);
	}


}

void SARDriver::forward_abr(Packet *c)
{
	if (cell_cnt_for_abr < (aal->get_reassembly().txtp)->nrm) {
		printf("SARDriver::forward_abr,c = 0x%x\n",c);

		cell_cnt_for_abr++;
		alwd_cell_cnt--;
		sar_q->enque(c);
	}	
	
	else {
		Packet *frm = aal->build_frm(c);
		printf("SAR:: sending frm, frm = 0x%x\n",frm);
		alwd_cell_cnt--;
		sar_q->enque(frm);
		cell_cnt_for_abr = 0;

		if (alwd_cell_cnt > 0) {
			cell_cnt_for_abr++;
			alwd_cell_cnt--;
			sar_q->enque(c);
		}

		else {
			printf("SARDRIVER::storing in abr buffer\n");
			aal->store_in_abr_buffer(c);
		}
	}
	
}


char *decode_bool(bool b) {

      if (b == true)
	return "true";

      else if (b == false)
        return "false";

      else
         return "unknown";

}














