#ifndef __sar_h__
#define __sar_h__

#include "connector.h"
#include "packet.h"
#include "atmqos.h"
#include "aal.h"
#include "atm.h"
#include "atmpvc.h"
#include "queue.h"

class SARDriver;
class SARTimer;

class SARTimer : public TimerHandler {
 public:
  SARTimer(SARDriver *s) : TimerHandler(), sar_(s) {}
  inline virtual void expire(Event *);
 protected:
  SARDriver *sar_;
};


class SARDriver : public Agent {

 public:
  SARDriver();
  SARDriver(packet_t);
  ~SARDriver() { printf("SAR::destructor\n"); delete sar_q; }
  void recv(Packet *p, Handler *h);
  void resume();
  void timeout();
  void start_timer();

 protected:

  void aal5_segmentation(Packet *, Handler *);
  int command(int, const char*const* );
  NsObject *find(Packet *p) { NsObject *n; return n;} //dummy function, must be changed later
  inline double txtime() {
    return (8. * CELLSIZE / bw_);
  }
  void process_abr(Packet *, Handler *);
  inline double next_time();
  void forward_abr(Packet *);
  void send_brm(Packet *, Handler *);
  
  Event intr_;
  PacketQueue *sar_q;
  SARTimer timer_;

 private:
  bool atm;           
  AAL5 *aal;         
  bool atmswitch;
  int cell_cnt_for_abr;
  pvc_db *pvc_database;
  double bw_;
  double delay_;
  int off_atm_;
  int sent_cells;
  int alwd_cell_cnt;

  //statistics collection
  unsigned int rx_pkts;
  unsigned int tx_cells;
  unsigned int cells_per_pkt;
  
};

#endif
