/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 1996 Regents of the University of California.
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 * 	This product includes software developed by the MASH Research
 * 	Group at the University of California Berkeley.
 * 4. Neither the name of the University nor of the Research Group may be
 *    used to endorse or promote products derived from this software without
 *    specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char rcsid[] =
    "@(#) $Header: /nfs/jade/vint/CVSROOT/ns-2/classifier.cc,v 1.35 2000/09/14 18:19:25 haoboy Exp $";
#endif

#include <stdlib.h>
#include "config.h"
#include "classifier.h"
#include "packet.h"
#include <string>
#include "atm/atmpvc.h"
#include "atm/aal.h"


static class ClassifierClass : public TclClass {
public:
	ClassifierClass() : TclClass("Classifier") {}
	TclObject* create(int, const char*const*) {
		return (new Classifier());
	}
} class_classifier;


Classifier::Classifier() : 
	slot_(0), nslot_(0), maxslot_(-1), shift_(0), mask_(0xffffffff),
        atm(false), atmswitch(false), cell_cnt_for_abr(-1)
{
	default_target_ = 0;
	aal = new AAL5;

	bind("offset_", &offset_);
	bind("shift_", &shift_);
	bind("mask_", &mask_);

	rx_cells = 0;
	tx_pkts = 0;
}

int Classifier::classify(Packet *p)
{
	return (mshift(*((int*) p->access(offset_))));
}

Classifier::~Classifier()
{
	delete [] slot_;
	delete aal;
}

void Classifier::alloc(int slot)
{
	NsObject** old = slot_;
	int n = nslot_;
	if (old == 0)
		nslot_ = 32;
	while (nslot_ <= slot)
		nslot_ <<= 1;
	slot_ = new NsObject*[nslot_];
	memset(slot_, 0, nslot_ * sizeof(NsObject*));
	for (int i = 0; i < n; ++i)
		slot_[i] = old[i];
	delete [] old;
}


void Classifier::install(int slot, NsObject* p)
{
	if (slot >= nslot_)
		alloc(slot);
	slot_[slot] = p;
	if (slot >= maxslot_)
		maxslot_ = slot;
}

void Classifier::clear(int slot)
{
	slot_[slot] = 0;
	if (slot == maxslot_) {
		while (--maxslot_ >= 0 && slot_[maxslot_] == 0)
			;
	}
}

int Classifier::getnxt(NsObject *nullagent)
{
	int i;
	for (i=0; i < nslot_; i++)
		if (slot_[i]==0 || slot_[i]==nullagent)
			return i;
	i=nslot_;
	alloc(nslot_);
	return i;
}

/*
 * objects only ever see "packet" events, which come either
 * from an incoming link or a local agent (i.e., packet source).
 */
void Classifier::recv(Packet* p, Handler*h)
{

        hdr_atm *atmh;
	NsObject* node = find(p);
	if (node == NULL) {
		/*
		 * XXX this should be "dropped" somehow.  Right now,
		 * these events aren't traced.
		 */
		Packet::free(p);
		return;
	}

       //ATM extension, CNRG, UNH

        string d = string("dst_aal");
        string s = string("src_aal");

        if(p->is_atm())
                {
                        atmh = hdr_atm::access(p);
			if ((atmh->vpi == 0) && (atmh->vci == 0))
				{
			               printf("Clsfer::dropping packet, id = %s, vpi/vci=%d/%d, time = %e\n",name(),atmh->vpi,atmh->vci,Scheduler::instance().clock());
					Packet::free(p);
					return;
				}

                }

        if ((atm) && (p->is_atm()) && (!atmswitch) && (aal->aal_type() == d))
                {
			rx_cells++;
			aal5_reassembly(p,node,h);
                }


        else if ((atm) && (atmswitch) && (p->is_atm()))
                {
                 
                        switching(p,node,h);
                }

        else if ((atm) && (p->is_atm()) && (!atmswitch) && (aal->aal_type() == s)) {
		atmh = hdr_atm::access(p);
		if ((atmh->type == ATM_RM) && ((atmh->payload)->at(8) == '1')) {
			printf("Classifier::received BRM cell, I am %s\n",name());
			aal->adjust_source_rate(p);
		}

		else {
			node->recv(p,h);
		}
	}
        else {
		hdr_cmn *th = hdr_cmn::access(p);
		node->recv(p,h);                
	}

}


/*
 * perform the mapping from packet to object
 * perform upcall if no mapping
 */

NsObject* Classifier::find(Packet* p)
{
	NsObject* node = NULL;
	int cl = classify(p);
	if (cl < 0 || cl >= nslot_ || (node = slot_[cl]) == 0) { 
		if (default_target_) 
			return default_target_;
		/*
		 * Sigh.  Can't pass the pkt out to tcl because it's
		 * not an object.
		 */
		Tcl::instance().evalf("%s no-slot %ld", name(), cl);
		if (cl == TWICE) {
			/*
			 * Try again.  Maybe callback patched up the table.
			 */
			cl = classify(p);
			if (cl < 0 || cl >= nslot_ || (node = slot_[cl]) == 0)
				return (NULL);
		}
	}
	return (node);
}

int Classifier::install_next(NsObject *node) {
        int slot = maxslot_ + 1;
        install(slot, node);
        return (slot);
}



int Classifier::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if(argc == 2) {

		if (strcmp(argv[1], "atm_statistics") == 0) {
			printf("   %8d             %12d           \n",rx_cells,tx_pkts);
		  return TCL_OK;
		}
			
                if (strcmp(argv[1], "defaulttarget") == 0) {
                        if (default_target_ != 0)
                                tcl.result(default_target_->name());
                        return (TCL_OK);
                }

                if (strcmp(argv[1], "is_atmswitch") == 0) {
                        atmswitch = true;
                        return (TCL_OK);
                }

        } else if (argc == 3) {
		/*
		 * $classifier alloc-port nullagent
		 */
		if (strcmp(argv[1],"alloc-port") == 0) {
			int slot;
			NsObject* nullagent =
				(NsObject*)TclObject::lookup(argv[2]);
			slot = getnxt(nullagent);
			tcl.resultf("%u",slot);
			return(TCL_OK);
		}
		/*
		 * $classifier clear $slot
		 */
		if (strcmp(argv[1], "clear") == 0) {
			int slot = atoi(argv[2]);
			clear(slot);
			return (TCL_OK);
		}
		/*
		 * $classifier installNext $node
		 */
		if (strcmp(argv[1], "installNext") == 0) {
			int slot = maxslot_ + 1;
			NsObject* node = (NsObject*)TclObject::lookup(argv[2]);
			if (node == NULL) {
				tcl.resultf("Classifier::installNext attempt "
		    "to install non-object %s into classifier", argv[2]);
				return TCL_ERROR;
			};
			install(slot, node);
			tcl.resultf("%u", slot);
			return TCL_OK;
		}
		/*
		 * $classifier slot snum
		 * returns the name of the object in slot # snum
		 */
		if (strcmp(argv[1], "slot") == 0) {
			int slot = atoi(argv[2]);
			if (slot >= 0 && slot < nslot_ && slot_[slot] != NULL) {
				tcl.resultf("%s", slot_[slot]->name());
				return TCL_OK;
			}
			tcl.resultf("Classifier: no object at slot %d", slot);
			return (TCL_ERROR);
		}
		/*
		 * $classifier findslot $node
		 * finds the slot containing $node
		 */
		if (strcmp(argv[1], "findslot") == 0) {
			int slot = 0;
			NsObject* node = (NsObject*)TclObject::lookup(argv[2]);
			if (node == NULL) {
				return (TCL_ERROR);
			}
			while (slot < nslot_) {
				if (strcmp(slot_[slot]->name(), argv[2]) == 0) {
					tcl.resultf("%u", slot);
					return (TCL_OK);
				}
				slot++;
			}
			tcl.result("-1");
			return (TCL_OK);
		}
		if (strcmp(argv[1], "defaulttarget") == 0) {
			default_target_=(NsObject*)TclObject::lookup(argv[2]);
			if (default_target_ == 0)
				return TCL_ERROR;
			return TCL_OK;
		}

                if(strcmp(argv[1], "configure_atmswitch") == 0) {
                        pvc_database = (pvc_db*)TclObject::lookup(argv[2]);
                        if(pvc_database == NULL) {
                                tcl.resultf("Classifier::invalid pvc_database va
lue %s\n",argv[2]);
                                return TCL_ERROR;
                        }
                        atmswitch = true;
			atm = true;
                        return TCL_OK;
                }

	} else if (argc == 4) {
		/*
		 * $classifier install $slot $node
		 */
		if (strcmp(argv[1], "install") == 0) {
			int slot = atoi(argv[2]);
			NsObject* node = (NsObject*)TclObject::lookup(argv[3]);
			install(slot, node);
			return (TCL_OK);
		}
	} else if (argc == 7)  {   /* ATM extn - CNRG, UNH */
                                // enabling IP over ATM support
                if(strcmp(argv[1], "attach_aal5") == 0) {
                        atm = true;
                        Reassembly *r = new Reassembly;
                        r->vpi = atoi(argv[2]);
                        r->vci = atoi(argv[3]);
                        r->rxtp = (AtmTrafprm*)TclObject::lookup(argv[4]);
                        r->txtp = (AtmTrafprm*)TclObject::lookup(argv[5]);

                        string type = string(argv[6]);
                        aal->set_parameters(r, type);

                        /* if abr, RM cells must be generated after every
                           cell_cnt_for_abr = (nrm - 1), hence initialize
                           cell_cnt_for_abr to 0.  Note this also indicates
                           whether traffic class is abr or not. */

                        if((r->txtp)->traffic_class == ABR)
                                cell_cnt_for_abr = 0;

                        return TCL_OK;
                }
        }


	return (NsObject::command(argc, argv));
}



//destination node reassembly
void Classifier::aal5_reassembly(Packet *p, NsObject *node, Handler *h) {

        hdr_atm *atmh = hdr_atm::access(p);
	Packet *temp;
	hdr_cmn *th;

        /* if this is the first cell of the frame, note down
           the vpi/vci marker */
	
        int vp = aal->get_reassembly().vpi;
        int vc = aal->get_reassembly().vci;


        if((atmh->vpi == vp) && (atmh->vci == vc)) {

                int t = atmh->type;

		switch (t) {
			
			// EFCI bit set by switch
		case EFCI:
                        delete atmh->payload;
			Packet::free(p);
			aal->set_efci();
			break;
			
		case EFCI_LAST_CELL:
		case EFCI_LAST_UNKNOWN:
			th = hdr_cmn::access(atmh->packet);
//			printf("Classifier::reassembly, sending packet to agent, vpi/vci=%d/%d, orig pkt_id = %d \n",atmh->vpi,atmh->vci,th->uid());

			tx_pkts++;
			aal->set_efci();
                        temp = atmh->packet;
                        delete atmh->payload;
	                Packet::free(p);
                        node->recv(temp,h);
			break;
			
		case EFCI_FIVE:
		case LAST_CELL:
			th = hdr_cmn::access(atmh->packet);
                        temp = atmh->packet;
//			printf("Classifier::reassembly, sending packet to agent, vpi/vci=%d/%d, orig pkt_id = %d \n",atmh->vpi,atmh->vci,th->uid());
			tx_pkts++;
                        delete atmh->payload;
	                Packet::free(p);
                        node->recv(temp,h);
			break;
			
			// This packet is not going to be sent to target
			// Only the reassembled wd. be.  So, free some
			// resources
			
		case ATM_NONE:
		case ATM_FOUR:
                        delete atmh->payload;
			Packet::free(p);
			break;
			
			// this is an RM cell
		case ATM_RM:
                        treat_rm_cell(p,node,h);
			break;
			
		default:
			printf("CLASSIFIER::error in ATM type\n");
			return;
		}
	}

       else {
           printf("Classifier::dropping packet, vpi/vci did not match\n");
	   Packet::free(p);
       }
        

	return;
}

//ATM EXTENSION, CNRG, UNH
//atm switch forwarding
void Classifier::switching(Packet *p, NsObject *node, Handler *h) {

        hdr_atm *atmh = hdr_atm::access(p);
        int cmn = (hdr_cmn::access(atmh->packet))->uid();
	int atmp = (hdr_cmn::access(p))->uid();

        string s = string(atmh->switcher_);

        if (s != name())
                {

                        PVC *pvc = pvc_database->get_outgoing_pvc(atmh->vpi,atmh
->vci);

                        /* this should never happen */
                        if (pvc == NULL) {
                                atmh->vpi = -1;
                                atmh->vci = -1;

                        } else {
                                atmh->vpi = pvc->get_vpi();
                                atmh->vci = pvc->get_vci();
				strcpy(atmh->switcher_,((string)name()).c_str());

                        }

                }

        node->recv(p,h);
}

//ATM Extension, CNRG, UNH
//function to support ABR feedback generator

void Classifier::treat_rm_cell(Packet *p,NsObject *node, Handler *h) {

        hdr_atm *atmh = hdr_atm::access(p);

	cell_cnt++;
 
        //this is an FRM cell
        if ((atmh->payload)->at(8) == '0') {

		printf ("CLASSIFIER::received FRM CELL %d\n", cell_cnt);
                (atmh->payload)->at(8) = '1';
                if (aal->get_efci())
                        atmh->payload->at(10) = '1';

                //this is to send back the brm to the source
                // aal->reverse_addresses_for_brm(p);

                NsObject *n = find(p);
		n->recv(p,h);

        }

        //this is a BRM cell
        else if(atmh->payload->at(8) == '1') {
                NsObject *n = find(p);
		n->recv(p,h);
        }

	else
		printf("Classifier::Regular cell  %d\n",cell_cnt);

}





